// *******************************************************************
// Copyright (C) 2009 Telos Systems, TLS Corp.
// All rights reserved.
//
// No part of this source code or associated documentation may be 
// reproduced or utilized in any form or by any means, electronic or 
// mechanical, without permission in writing from TLS Corp.
// -------------------------------------------------------------------
//      File: APSend.h
// Description: See class description.
//    Author: Ioan L. Rus
//      Date: Nov 18, 2009
// Revisions: 
// -------------------------------------------------------------------
// INI Date        Description
// --- ----------- ---------------------------------------------------
// 
// *******************************************************************
#include <memory.h>
#include <string.h>
#include "Msg2.h"


#ifdef WIN32
  #include <winsock2.h>  // for hton#() and ntoh#() functions
  #pragma comment(lib,"ws2_32")
#else
#endif // Win32


static bool s_GetLong(const char *pData,unsigned long &lData,unsigned long &ulVal,unsigned long cbData)
{
  // The operation would go past the end of buffer!
  if((lData+sizeof(unsigned long)) > cbData)
    return false;

  memcpy(&ulVal,pData+lData,sizeof(unsigned long));
  lData+=sizeof(unsigned long);

  ulVal=ntohl(ulVal);

  return true;
}

static bool s_GetShort(const char *pData,unsigned long &lData,unsigned short &usVal,unsigned long cbData)
{
  // The operation would go past the end of buffer!
  if((lData+sizeof(unsigned short)) > cbData)
    return false;

  memcpy(&usVal,pData+lData,sizeof(unsigned short));
  lData+=sizeof(unsigned short);

  usVal=ntohs(usVal);

  return true;
}

static void s_SetLong(char *pData,CMSG2_SIZE &lData,unsigned long ulVal)
{
  ulVal=htonl(ulVal);
  memcpy(pData+lData,&ulVal,sizeof(long));
  lData+=sizeof(long);
}

static void s_SetShort(char *pData,CMSG2_SIZE &lData,unsigned short usVal)
{
  usVal=htons(usVal);
  memcpy(pData+lData,&usVal,sizeof(unsigned short));
  lData+=sizeof(unsigned short);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// Creates a CMsg2 object either empty or by specifying another CMsg2
// or byte stream representation or a CMsg2.
CMsg2::CMsg2(const char *pData,    // Byte-stream packed message used as the source.
             unsigned long cbData, // Number of bytes in the data buffer.
             unsigned int nGrowBy  // Determines the number of items added to the
                                   // message when there is no space for the new item
                                   // being added.  A large value wastes memory while
                                   // too small of a value is slow.  The default value
                                   // is 16.
             )
  :
  m_nSize(0),
  m_nGrowBy(nGrowBy),
  m_ulMessageId(0),
  m_nIndex(0xFFFF),
  m_nCount(0),
  m_pNodeArray(0)
{
  SetData(pData,cbData);
}

CMsg2::CMsg2(unsigned int nInitialSize, // Specifies the number of items that can be
                                        // stored in this message.  If nGrowBy is not
                                        // zero the message will grow past nInitialSize
                                        // if more items are added to the message.
             unsigned int nGrowBy)
  :
  m_nSize(nInitialSize),
  m_nGrowBy(nGrowBy),
  m_ulMessageId(0),
  m_nIndex(0xFFFF),
  m_nCount(0)
{
  m_pNodeArray=new Node[m_nSize];
}

CMsg2::CMsg2(const CMsg2 &msg // Source CMsg2 to copy.
             )
  :
  m_nCount(0),
  m_nIndex(0xFFFF),
  m_pNodeArray(0)
{
  *this=msg;
}

// Desc: Return the version ID for the CMsg2 class.
// Remarks: The version ID indicates message content compatibility.
// The version ID is used to tag byte-stream representations
// of messages before they are sent over the network. If the receiver's
// version ID matches it means that the receiver can safely parse the
// packet contents.
// Returns: The version ID for the CMsg2 class.
unsigned char CMsg2::GetVer()
{
  // CMsg2 can use 2-byte or 4-byte size fields.
  // (see USE_WORD_SIZE in the header file for 
  // more details).
  //
  // The most significant bit of the version id 
  // indicates if 2-byte or 4-byte sizes are used.
  // MSb=0 -> 2-byte sizes
  // MSb=1 -> 4-byte sizes
  //
  // The other 7 bits are used for the version.
  //
  unsigned char nVersion=1;

#ifdef USE_WORD_SIZE
  nVersion&=~0x80;
#else
  nVersion|=0x80;
#endif // USE_WORD_SIZE

  return nVersion;
}

// Copy the contents of the specified CMsg2 to this object.
// The existing contents of this message are discarded prior
// to Copy.
void CMsg2::Copy(const CMsg2 *pMsg // Pointer to source CMsg2 to copy.
                 )
{
  if(pMsg!=0)
	  *this=*pMsg;
}

// Add the contents of the specified CMsg2 object to this object.
// Remarks:
// Add the contents of the specified CMsg2 object to this object.
// The existing contents of this message are not discarded prior
// to Copy.  Items with the same tag from the second message will
// overwrite items in this message.  The message Id is not changed.
CMsg2 &CMsg2::operator+=(const CMsg2 &msg  // Source CMsg2 to copy.
                         )
{
  for(unsigned int i=0; i<msg.m_nCount; i++)
  {
    Node *pNode=&msg.m_pNodeArray[i];
    switch(pNode->m_eType)
    {
    case TypByte:
      SetByte(pNode->m_ulName,(char)(pNode->m_Value.m_lNumber & 0xFF));
      break;

    case TypWord:
      SetWord(pNode->m_ulName,(short)(pNode->m_Value.m_lNumber & 0xFFFF));
      break;

    case TypDword:
      SetDword(pNode->m_ulName,pNode->m_Value.m_lNumber);
      break;

    case TypString:
      SetString(pNode->m_ulName,pNode->m_Value.m_pString);
      break;

    case TypByteArray:
      SetByteArray(pNode->m_ulName,pNode->m_nArrayCount,pNode->m_Value.m_pByteArray);
      break;

    case TypWordArray:
      SetWordArray(pNode->m_ulName,pNode->m_nArrayCount,pNode->m_Value.m_pWordArray);
      break;

    case TypDwordArray:
      SetDwordArray(pNode->m_ulName,pNode->m_nArrayCount,pNode->m_Value.m_pDwordArray);
      break;

    case TypMsg:
      SetMsg(pNode->m_ulName,pNode->m_Value.m_pMsg);
      break;
    }
  }
  return *this;
}

// Copy the contents of the specified CMsg2 to this object.
// Remarks:
// Copy the contents of the specified CMsg2 to this object.
// The existing contents of this message are discarded prior
// to Copy.
CMsg2 &CMsg2::operator=(const CMsg2 &msg  // Source CMsg2 to copy.
                        )
{
  Empty();

  if(m_pNodeArray!=0)
    delete [] m_pNodeArray;

  m_nSize=msg.m_nSize;
  m_nGrowBy=msg.m_nGrowBy;
  m_ulMessageId=msg.m_ulMessageId;

  m_pNodeArray=new Node[m_nSize];

  *this+=msg;

  return *this;
}

// Destructor.
CMsg2::~CMsg2()
{
  Empty();

  if(m_pNodeArray!=0)
    delete [] m_pNodeArray;
}

// Remove all items from the message.
void CMsg2::Empty()
{
  for(unsigned int i=0; i<m_nCount; i++)
  {
    Node *pNode=&m_pNodeArray[i];
    pNode->DeleteValue();
    pNode->m_ulName=0;
  }

  m_nCount=0;
}

// Return the byte count needed to store this message.
// Returns: The byte count needed to store this message.
CMSG2_SIZE CMsg2::GetDataSize() const
{
  CMSG2_SIZE lData=sizeof(long);  // Message id
  
#ifdef USE_WORD_SIZE
  lData+=sizeof(short);  // Count of items
#else  // USE_WORD_SIZE
  lData+=sizeof(long);  // Count of items
#endif // USE_WORD_SIZE

  for(unsigned int i=0; i<m_nCount; i++)
  {
    Node *pNode=&m_pNodeArray[i];

    if(pNode->m_eType!=TypInvalid)
    {
      lData+=sizeof(long)+sizeof(unsigned char); // Name+ItemType
      if(pNode->m_eType==TypDword)
        lData+=sizeof(long);  // Dword value
      else if(pNode->m_eType==TypByte)
        lData+=sizeof(char);  // Byte value
      else if(pNode->m_eType==TypWord)
        lData+=sizeof(short); // Word value
      else
#ifdef USE_WORD_SIZE
        lData+=sizeof(short);  // Array count
#else  // USE_WORD_SIZE
        lData+=sizeof(long);  // Array count
#endif // USE_WORD_SIZE
      switch(pNode->m_eType)
      {
      case TypString:
        lData+=pNode->m_nArrayCount*sizeof(char);
        break;
      case TypByteArray:
        lData+=pNode->m_nArrayCount*sizeof(unsigned char);
        break;
      case TypWordArray:
        lData+=pNode->m_nArrayCount*sizeof(unsigned short);
        break;
      case TypDwordArray:
        lData+=pNode->m_nArrayCount*sizeof(unsigned long);
        break;
      case TypMsg:
        lData+=pNode->m_Value.m_pMsg->GetDataSize();
        break;
      }
    }
  }

  return lData;
}

// Retrieve the byte-stream representation of this message.
// Remarks: The GetData() function retrieves the byte-stream
// representation of the message.  The message contents
// are serialized and written to the pData buffer.  Note
// that the buffer pointed to by pData must be at least
// GetDataSize() bytes long.  
//
// When a message is serialized there is no guarantee to
// the order of the items.  Items can be later identified
// by their unique tag.
//
// The USE_WORD_SIZE define controls how messages are
// converted to a byte stream.  Take a look at USE_WORD_SIZE
// for more details.
//
// Returns: Number of bytes used to store the message.
//
// See Also: GetDataSize(), SetData(), USE_WORD_SIZE
//
CMSG2_SIZE CMsg2::GetData(char *pData // Pointer to buffer receiving the
                                      // message contents.  This buffer
                                      // must be at least GetDataSize()
                                      // bytes long.
                          ) const
{
  unsigned int n;
  CMSG2_SIZE lData=0;

  s_SetLong(pData,lData,m_ulMessageId);
#ifdef USE_WORD_SIZE
  s_SetShort(pData,lData,m_nCount);
#else  // USE_WORD_SIZE
  s_SetLong(pData,lData,m_nCount);
#endif // USE_WORD_SIZE

  for(unsigned int i=0; i<m_nCount; i++)
  {
    Node *pNode=&m_pNodeArray[i];

    if(pNode->m_eType!=TypInvalid)
    {
      s_SetLong(pData,lData,pNode->m_ulName);

      pData[lData]=(char)pNode->m_eType;
      lData+=sizeof(unsigned char); // ItemType

      if(pNode->m_eType==TypDword)
      {
        s_SetLong(pData,lData,pNode->m_Value.m_lNumber);
      }
      else if(pNode->m_eType==TypWord)
      {
        s_SetShort(pData,lData,(short)(pNode->m_Value.m_lNumber & 0xFFFF));
      }
      else if(pNode->m_eType==TypByte)
      {
        pData[lData++]=(char)(pNode->m_Value.m_lNumber & 0xFF);
      }
      else if(pNode->m_eType==TypMsg)
      {
#ifdef USE_WORD_SIZE
        s_SetShort(pData,lData,(unsigned short)pNode->m_Value.m_pMsg->GetDataSize());
#else  // USE_WORD_SIZE
        s_SetLong(pData,lData,pNode->m_Value.m_pMsg->GetDataSize());
#endif // USE_WORD_SIZE
        lData+=pNode->m_Value.m_pMsg->GetData(pData+lData);
      }
      else  // Must be array
      {
#ifdef USE_WORD_SIZE
        s_SetShort(pData,lData,pNode->m_nArrayCount);
#else  // USE_WORD_SIZE
        s_SetLong(pData,lData,pNode->m_nArrayCount);
#endif // USE_WORD_SIZE

        switch(pNode->m_eType)
        {
        case TypString:
        case TypByteArray:
          if(pNode->m_nArrayCount>0)
          {
            memcpy(pData+lData,pNode->m_Value.m_pString,pNode->m_nArrayCount);
            lData+=pNode->m_nArrayCount*sizeof(char);
          }
          break;
        case TypWordArray:
          for(n=0; n<pNode->m_nArrayCount; n++)
          {
            s_SetShort(pData,lData,pNode->m_Value.m_pWordArray[n]);
          }
          break;
        case TypDwordArray:
          for(n=0; n<pNode->m_nArrayCount; n++)
          {
            s_SetLong(pData,lData,pNode->m_Value.m_pDwordArray[n]);
          }
          break;
        }
      }
    }
  }

  return lData;
}

// Parse a byte-stream into a CMsg2 object.
// Remarks: The SetData() method parses a byte-stream representation of
// a message into a CMsg2 object.
//
// Returns: Number of bytes parsed from buffer pointed to by pData.  The
// method returns zero if a valid CMsg2 could not be parsed from the
// contents of pData or if parsing the contents of pData would exceed
// the boundary specified by cbData.
//
// See Also: GetData(), USE_WORD_SIZE
CMSG2_SIZE CMsg2::SetData(const char *pData,   // Pointer to buffer holding the byte-stream
                                               // representation of the message.
                          unsigned long cbData // Number of bytes in the data buffer.
                          )
{
  unsigned int n;
  //CMSG2_SIZE lData=0;
  unsigned long lData=0;

  Empty();

  if(!s_GetLong(pData,lData,m_ulMessageId,cbData))
    goto failure_exit;

  // Figure out if the message is large enough
#ifdef USE_WORD_SIZE
  unsigned short nCount;
  if(!s_GetShort(pData,lData,nCount,cbData))
    goto failure_exit;
#else  // USE_WORD_SIZE
  unsigned long nCount;
  if(!s_GetLong(pData,lData,nCount,cbData))
    goto failure_exit;
#endif // USE_WORD_SIZE

  if(m_nSize<nCount)
  {
    m_nSize=nCount;
    if(m_pNodeArray!=0)
      delete [] m_pNodeArray;
    m_pNodeArray=new Node[m_nSize];
  }
  m_nCount=nCount;

  unsigned int i;
  for(i=0; i<m_nCount; i++)
  {
    Node *pNode=&m_pNodeArray[i];

    if(!s_GetLong(pData,lData,pNode->m_ulName,cbData))
      goto failure_exit;

    if((lData+sizeof(unsigned char)) > cbData)
      goto failure_exit;
    pNode->m_eType=(unsigned char)pData[lData];
    lData+=sizeof(unsigned char); // ItemType

    if(pNode->m_eType==TypDword)
    {
      unsigned long ulVal;
      if(!s_GetLong(pData,lData,ulVal,cbData))
        goto failure_exit;
      pNode->m_Value.m_lNumber=ulVal;
      pNode->m_nArrayCount=0;
    }
    else if(pNode->m_eType==TypWord)
    {
      unsigned short usVal;
      if(!s_GetShort(pData,lData,usVal,cbData))
        goto failure_exit;
      pNode->m_Value.m_lNumber=usVal;
      pNode->m_nArrayCount=0;
    }
    else if(pNode->m_eType==TypByte)
    {
      if((lData+sizeof(unsigned char)) > cbData)
        goto failure_exit;
      pNode->m_Value.m_lNumber=pData[lData++];
      pNode->m_nArrayCount=0;
    }
    else if(pNode->m_eType==TypMsg)
    {
#ifdef USE_WORD_SIZE
      unsigned short cbMsgSize;
      if(!s_GetShort(pData,lData,cbMsgSize,cbData))
        goto failure_exit;
#else  // USE_WORD_SIZE
      unsigned long  cbMsgSize;
      if(!s_GetLong(pData,lData,cbMsgSize,cbData))
        goto failure_exit;
#endif // USE_WORD_SIZE
      if((lData+cbMsgSize) > cbData)
        goto failure_exit;
      pNode->m_Value.m_pMsg=new CMsg2(pData+lData,cbMsgSize);
      lData+=cbMsgSize;
    }
    else  // Must be array
    {
#ifdef USE_WORD_SIZE
      if(!s_GetShort(pData,lData,pNode->m_nArrayCount,cbData))
        goto failure_exit;
#else  // USE_WORD_SIZE
      if(!s_GetLong(pData,lData,pNode->m_nArrayCount,cbData))
        goto failure_exit;
#endif // USE_WORD_SIZE

      if(pNode->m_nArrayCount!=0)
      {
        switch(pNode->m_eType)
        {
        case TypString:
          if(pNode->m_nArrayCount>0)
          {
            if((lData+(pNode->m_nArrayCount*sizeof(char))) > cbData)
              goto failure_exit;
            pNode->m_Value.m_pString=new char[pNode->m_nArrayCount];
            memcpy(pNode->m_Value.m_pString,pData+lData,pNode->m_nArrayCount);
            lData+=pNode->m_nArrayCount*sizeof(char);
          }
          else
          {
            // NULL string
            pNode->m_Value.m_pString=0;
          }
          break;

        case TypByteArray:
          if((lData+(pNode->m_nArrayCount*sizeof(unsigned char))) > cbData)
            goto failure_exit;
          pNode->m_Value.m_pByteArray=new unsigned char[pNode->m_nArrayCount];
          memcpy(pNode->m_Value.m_pByteArray,pData+lData,pNode->m_nArrayCount);
          lData+=pNode->m_nArrayCount*sizeof(unsigned char);
          break;

        case TypWordArray:
          pNode->m_Value.m_pWordArray=new unsigned short[pNode->m_nArrayCount];
          for(n=0; n<pNode->m_nArrayCount; n++)
          {
            if(!s_GetShort(pData,lData,pNode->m_Value.m_pWordArray[n],cbData))
              goto failure_exit;
          }
          break;
        case TypDwordArray:
          pNode->m_Value.m_pDwordArray=new unsigned long[pNode->m_nArrayCount];
          for(n=0; n<pNode->m_nArrayCount; n++)
          {
            if(!s_GetLong(pData,lData,pNode->m_Value.m_pDwordArray[n],cbData))
              goto failure_exit;
          }
          break;
        }
      }
    }
  }

  if(lData>CMSG2_SIZE_MAX)
    goto failure_exit;

  return (CMSG2_SIZE)lData;

// I know you don't like goto statements, but I don't 
// like the alternative.  Sure, exceptions would be better
// but they are not supported on some platforms.
failure_exit:
  Empty();
  return 0;
}

// Set the message Id.
// Remarks: This method sets the Id for the message.  The Id is
// used to identify this message's contents, source or destination.
// The Id can be any *unsigned long* number although it is
// strongly recommended that zero not be used.  The default value 
// for the message Id is zero
// and will remain zero if you forget to call SetId() for a
// particular message.  Seeing a message with a zero Id quickly
// indicates that no SetId() has been called on that message.
void CMsg2::SetId(unsigned long ulMessageTag // Message Id tag.
                  )
{
  m_ulMessageId=ulMessageTag;
}

// Get this message's Id.
// Returns: The message id.
unsigned long CMsg2::GetId() const
{
  return m_ulMessageId;
}

// Desc: Return the numeric tag for a given string or number.
// Remarks: There are two forms of the GetTag() method.
// The first form converts a name string to a tag value (4-byte number).
// Only the first four bytes of the string are processed.
// It is OK for the string to be shorter than four characters.
//
// The second converts a number plus an optional prefix to
// a tag value.  The prefix can be up to three characters long.
// If no prefix is used the maximum value for the number is
// 9999.  If a prefix is used the maximum value is reduced by an
// order of 10 for each character.  This means that a two character
// prefix would reduce the maximum value for the number to 99.
// It is OK to pass a NULL for the prefix if no prefix is used.
//
// Returns: Tag as an *unsigned long*.  Zero is returned if the name
// has no characters or if the number value passed in is too large
// for the given prefix.  See last example below.
//
// Examples: Here are some examples:
//
//      GetTag("ABCD")   returns 0x41424344
//      GetTag("AB")     returns 0x41420000
//      GetTag(NULL,1)   returns 0x30303031, same as GetTag("0001")
//      GetTag("AB",2)   returns 0x41423032, same as GetTag("AB02")
//      GetTag("AB",100) returns 0x00000000 because value of 100 does not fit.
//
unsigned long CMsg2::GetTag(const char *pszName // Name string to be converted to a 
                                                // tag value.  Only the first four
                                                // bytes of the string are processed.
                                                // It is OK for the string to be shorter
                                                // than four characters.
                            )
{
  unsigned long ulTag=0;
  const unsigned char *pszName1=(unsigned char *)pszName;

  if(pszName1!=0 && *pszName1!='\0')
  {
    ulTag=*pszName1++;

    ulTag<<=8;
    if(*pszName1!='\0')
      ulTag|=*pszName1++;

    ulTag<<=8;
    if(*pszName1!='\0')
      ulTag|=*pszName1++;

    ulTag<<=8;
    if(*pszName1!='\0')
      ulTag|=*pszName1;
  }

  return ulTag;
}

unsigned long CMsg2::GetTag(const char *pszPrefix, // Prefix string to be prepended to
                                                   // the uNumber when creating the tag.
                                                   // A maximum of 3 characters may be
                                                   // used for prefix.  NULL can be used
                                                   // in place of pszPrefix if no prefix
                                                   // is used.
                            unsigned int uIndex    // Number appended to the optional
                                                   // pszPrefix string when creating the
                                                   // tag.
                            )
{
  unsigned long ulTag=0;
  unsigned int uMaxNumber=1000;

  if(pszPrefix!=0)
  {
    const unsigned char *pszName=(unsigned char *)pszPrefix;

    for(int i=0; i<4 && *pszName!='\0'; i++)
    {
      ulTag|=*pszName++;
      if(i<3)
        ulTag<<=8;
      uMaxNumber/=10;
    }
  }

  if(uIndex>=(uMaxNumber*10))
  {
    // Can't encode number! Number too large for space left.
    return 0;
  }

  if(uMaxNumber>0)
  {
    for(;;)
    {
      ulTag|='0'+(unsigned char)((uIndex/uMaxNumber) % 10);
      uMaxNumber/=10;
      if(uMaxNumber>0)
        ulTag<<=8;
      else
        break;
    }
  }

  return ulTag;
}


// Converts a numeric tag back to a string. 
// Returns: A pointer to the buffer pointed to by pszName.
char *CMsg2::GetName(unsigned long ulTag, // Numeric tag
                     char *pszName        // Pointer to string buffer to hold 
                                          // tag name.  The buffer size must be at
                                          // least five bytes.
                     )
{
  unsigned char *pszName1=(unsigned char *)pszName;

  pszName1[4]='\0';
  pszName1[3]=(unsigned char)(ulTag&0xFF);
  ulTag>>=8;
  pszName1[2]=(unsigned char)(ulTag&0xFF);
  ulTag>>=8;
  pszName1[1]=(unsigned char)(ulTag&0xFF);
  ulTag>>=8;
  pszName1[0]=(unsigned char)(ulTag&0xFF);

  return pszName;
}

// Set a byte value.
// Returns: true on success, false on failure
bool CMsg2::SetByte(unsigned long ulTag, // Numeric tag
                    unsigned char ucNum // Value
                    )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypByte;
  pNode->m_Value.m_lNumber=ucNum;
  return true;
}

// Set a word (2-byte) value.
// Returns: true on success, false on failure
bool CMsg2::SetWord(unsigned long ulTag,  // Numeric tag
                    unsigned short usNum  // Value
                    )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypWord;
  pNode->m_Value.m_lNumber=usNum;
  return true;
}

// Set a 32 bit number.
// Returns: true on success, false on failure
bool CMsg2::SetDword(unsigned long ulTag, // Numeric tag
                     unsigned long ulNum // Value
                     )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypDword;
  pNode->m_Value.m_lNumber=ulNum;
  return true;
}

// Set a boolean value.
// Returns: true on success, false on failure
bool CMsg2::SetBool(unsigned long ulTag, // Numeric tag
                    bool bValue // Value
                    )
{
  return SetByte(ulTag,bValue ? 1 : 0);
}

// Set a string value.
// Returns: true on success, false on failure
bool CMsg2::SetString(unsigned long ulTag, // Numeric tag
                      const char *pszValue // Pointer to string
                      )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypString;
  if(pszValue!=0)
  {
    pNode->m_nArrayCount=(CMSG2_SIZE)strlen(pszValue)+1;
    pNode->m_Value.m_pString=new char[pNode->m_nArrayCount];
    memcpy(pNode->m_Value.m_pString,pszValue,pNode->m_nArrayCount);
  }
  else
  {
    // NULL string value
    pNode->m_nArrayCount=0;
    pNode->m_Value.m_pString=0;
  }
  return true;
}

// Return a string value.
// Returns: Pointer to string on success, NULL on failure.
const char *CMsg2::GetString(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypString)
  {
    // Not a string
    return 0;
  }

  return pNode->m_Value.m_pString;
}

// Return a byte value.
// Returns: Item value or zero if the item is not found or is not a byte type.
unsigned char CMsg2::GetByte(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypByte)
  {
    // Not a number
    return 0;
  }

  return (unsigned char)(pNode->m_Value.m_lNumber & 0xFF);
}

// Return a word value.
// Returns: Item value or zero if the item is not found or is not a word type.
unsigned short CMsg2::GetWord(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypWord)
  {
    // Not a number
    return 0;
  }

  return (unsigned short)(pNode->m_Value.m_lNumber & 0xFFFF);
}

// Return a 32-bit value.
// Returns: Item value or zero if the item is not found or is not a 32-bit number type.
unsigned long CMsg2::GetDword(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypDword)
  {
    // Not a 32-bit number
    return 0;
  }

  return pNode->m_Value.m_lNumber;
}


// Return any numeric value (byte, word, or dword).
// Returns: Item value or zero if the item is not found or is not a number type.
unsigned long CMsg2::GetNumber(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypByte && pNode->m_eType!=TypWord && pNode->m_eType!=TypDword)
  {
    // Not a number
    return 0;
  }

  return pNode->m_Value.m_lNumber;
}


// Return a bool value.
// Returns: Item value or false if the item is not found or is not a byte type.
bool CMsg2::GetBool(unsigned long ulTag /*Numeric tag*/)
{
  return GetByte(ulTag)!=0;
}

// Write a nested message inside this message.
// Returns: true on success, false on failure.
bool CMsg2::SetMsg(unsigned long ulTag /*Numeric tag*/,
                   const CMsg2 *pMsg // Pointer to CMsg2 to copy
                   )
{
  if(pMsg==0)
    return false;

  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypMsg;
  // Make a copy of the message
  pNode->m_Value.m_pMsg=new CMsg2(*pMsg);
  return true;
}

// Return a pointer to a nested message.
// Returns: Pointer to nested message of NULL if items not found or is not a message type.
CMsg2 *CMsg2::GetMsg(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypMsg)
  {
    // Not a number
    return 0;
  }

  return pNode->m_Value.m_pMsg;
}

// Set a byte array.
// Returns: true on success, false on error.
bool CMsg2::SetByteArray(unsigned long ulTag, // Item tag
                         unsigned int nItems, // Number of items for the array
                         unsigned char *pData // Pointer to array data to copy
                                              // into message.  If this parameter
                                              // is NULL no data is copied but
                                              // all array items are set to zero.
                                              // Array value can be set later using
                                              // the SetByteArrayItem() method.
                         )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypByteArray;
  pNode->m_nArrayCount=(CMSG2_SIZE)nItems;
  pNode->m_Value.m_pByteArray=new unsigned char[pNode->m_nArrayCount];
  if(pData!=0)
    memcpy(pNode->m_Value.m_pByteArray,pData,pNode->m_nArrayCount*sizeof(unsigned char));
  else
    memset(pNode->m_Value.m_pByteArray,0,pNode->m_nArrayCount*sizeof(unsigned char));
  return true;
}

// Set a word array.
// Returns: true on success, false on error.
bool CMsg2::SetWordArray(unsigned long ulTag,   // Item tag
                         unsigned int nItems,   // Number of items for the array
                         unsigned short *pData  // Pointer to array data to copy
                                              // into message.  If this parameter
                                              // is NULL no data is copied but
                                              // all array items are set to zero.
                                              // Array value can be set later using
                                              // the SetWordArrayItem() method.
                         )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypWordArray;
  pNode->m_nArrayCount=(CMSG2_SIZE)nItems;
  pNode->m_Value.m_pWordArray=new unsigned short[pNode->m_nArrayCount];
  if(pData!=0)
    memcpy(pNode->m_Value.m_pWordArray,pData,pNode->m_nArrayCount*sizeof(unsigned short));
  else
    memset(pNode->m_Value.m_pWordArray,0,pNode->m_nArrayCount*sizeof(unsigned short));
  return true;
}

// Set a dword array.
// Returns: true on success, false on error.
bool CMsg2::SetDwordArray(unsigned long ulTag,  // Item tag
                          unsigned int nItems,  // Number of items for the array
                          unsigned long *pData  // Pointer to array data to copy
                                              // into message.  If this parameter
                                              // is NULL no data is copied but
                                              // all array items are set to zero.
                                              // Array value can be set later using
                                              // the SetDwordArrayItem() method.
                          )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulTag);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypDwordArray;
  pNode->m_nArrayCount=(CMSG2_SIZE)nItems;
  pNode->m_Value.m_pDwordArray=new unsigned long[pNode->m_nArrayCount];
  if(pData!=0)
    memcpy(pNode->m_Value.m_pDwordArray,pData,pNode->m_nArrayCount*sizeof(unsigned long));
  else
    memset(pNode->m_Value.m_pDwordArray,0,pNode->m_nArrayCount*sizeof(unsigned long));
  return true;
}

// Set a value into an existing byte array.
// Returns: true on success, false on error.
bool CMsg2::SetByteArrayItem(unsigned long ulTag, // Item tag for the array.
                             unsigned int nIndex, // Index into the array.
                             unsigned char bData  // Data to set.
                             )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    return false;
  }

  if(pNode->m_eType!=TypByteArray)
  {
    return false;
  }

  if(pNode->m_nArrayCount<=nIndex)
  {
    return false;
  }

  pNode->m_Value.m_pByteArray[nIndex]=bData;

  return true;
}

// Set a value into an existing word array.
// Returns: true on success, false on error.
bool CMsg2::SetWordArrayItem(unsigned long ulTag,   // Item tag for the array.
                             unsigned int nIndex,   // Index into the array.
                             unsigned short usData  // Data to set.
                             )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    return false;
  }

  if(pNode->m_eType!=TypWordArray)
  {
    return false;
  }

  if(pNode->m_nArrayCount<=nIndex)
  {
    return false;
  }

  pNode->m_Value.m_pWordArray[nIndex]=usData;

  return true;
}

// Set a value into an existing dword array.
// Returns: true on success, false on error.
bool CMsg2::SetDwordArrayItem(unsigned long ulTag,  // Item tag for the array.
                              unsigned int nIndex,  // Index into the array.
                              unsigned long ulData  // Data to set.
                              )
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    return false;
  }

  if(pNode->m_eType!=TypDwordArray)
  {
    return false;
  }

  if(pNode->m_nArrayCount<=nIndex)
  {
    return false;
  }

  pNode->m_Value.m_pDwordArray[nIndex]=ulData;

  return true;
}

// Return number of items in array.
// Returns: Number of items in array.
CMSG2_SIZE CMsg2::GetArrayCount(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    return 0;
  }

  return pNode->m_nArrayCount;
}

// Return a pointer to a byte array.
// Returns: Pointer to array or NULL on error.
unsigned char *CMsg2::GetByteArray(unsigned long ulTag /* Numeric tag */)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypByteArray)
  {
    // Not a byte array
    return 0;
  }

  return pNode->m_Value.m_pByteArray;
}

short *CMsg2::GetShortArray(const char *pszName)
{
  return (short *)GetWordArray(pszName);
}

long *CMsg2::GetLongArray (const char *pszName)
{
  return (long *)GetDwordArray(pszName);
}

// Return a pointer to a word array.
// Returns: Pointer to array or NULL on error.
unsigned short *CMsg2::GetWordArray(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypWordArray)
  {
    // Not a word array
    return 0;
  }

  return pNode->m_Value.m_pWordArray;
}

// Return a pointer to a dword array.
// Returns: Pointer to array or NULL on error.
unsigned long  *CMsg2::GetDwordArray(unsigned long ulTag /* Numeric tag */)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
  {
    // Not found
    return 0;
  }

  if(pNode->m_eType!=TypDwordArray)
  {
    // Not a dword array
    return 0;
  }

  return pNode->m_Value.m_pDwordArray;
}

// Tests if an item with the given tag is defined in the message.
// Returns: true if item exists, otherwise false.
bool CMsg2::Exists(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  return(FindItem(ulTag)!=0);
}

// Removes the item with the given tag.
// Returns: true on success, false otherwise.
bool CMsg2::Remove(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
    return false;

  pNode->DeleteValue();
  pNode->m_ulName=0;

  int nCountToMove=m_nCount-m_nIndex-1;
  m_nCount--;

  if(nCountToMove>0)
    memmove(pNode,pNode+1,nCountToMove*sizeof(Node));

  return true;
}

// Returns the number of items in the message.
// Returns: A count of items in the message.
// See Also: GetItemTag(), GetItemName(), GetItemType()
unsigned int CMsg2::GetItemCount() const
{
  return m_nCount;
}

// Returns the tag of an item by index.
// Remarks: The GetItemTag() method can be used together with
// GetItemCount() and GetItemType() to explore the contents
// of an unknown message.  See the *DumpTextMsg* function
// in the *MessageDump* utility program for an example.
// Returns: true on success, false on failure.
// See Also: GetItemCount(), GetItemName(), GetItemType()
bool CMsg2::GetItemTag(unsigned int nIndex, // Item index
                       unsigned long &ulTag // Reference which will receive the item tag
                       ) const
{
  if(nIndex>=m_nCount)
    return false;

  ulTag=m_pNodeArray[nIndex].m_ulName;
  return true;
}

// Returns printable item name.
// Remarks: This method returns the name for an item by index.
// Note that only printable characters are returned by this
// method.  This means that if the item tag contains non-printable
// characters this method will not be able to return a valid name
// for the item.  Use GetItemTag() in this case.
// Returns: true on success, false on failure.
// See Also: GetItemCount(), GetItemTag(), GetItemType()
bool CMsg2::GetItemName(unsigned int nIndex, // Item index
                        char *pszName // Pointer to buffer that will receive 
                                      // the item name. Buffer must be at 
                                      // least 5 bytes long.
                        ) const
{
  unsigned long ulItemTag=0;

  if(!GetItemTag(nIndex,ulItemTag))
    return false;

  GetName(ulItemTag,pszName);

  for(int i=0; i<4; i++)
  {
    if(!isprint(pszName[i]))
    {
      pszName[i]='\0';
      break;
    }
  }

  return true;
}

// Return the type of an item.
// Returns: The item type.
// See Also: GetItemCount(), GetItemTag(), GetItemName()
int  CMsg2::GetItemType(unsigned long ulTag /*Numeric tag*/)
{
  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulTag);
  if(pNode==0)
    return TypInvalid;

  return pNode->m_eType;
}

CMsg2::Node *CMsg2::FindItem(unsigned long ulTag)
{
  // Zero item case
  if(m_pNodeArray==0 || m_nCount==0)
    return 0;

  // Single item case
  if(m_nCount==1)
  {
    m_nIndex=0;
    if(m_pNodeArray[0].m_ulName==ulTag)
    {
      return &m_pNodeArray[0];
    }
    else
    {
      return 0;
    }
  }

  // Multiple item case
  if(m_nIndex>=m_nCount)
    m_nIndex=0;

  unsigned int nLastItem=m_nIndex;

  do
  {
    if(m_pNodeArray[m_nIndex].m_ulName==ulTag)
      return &m_pNodeArray[m_nIndex];
    m_nIndex++;
    m_nIndex%=m_nCount;
  } while(m_nIndex!=nLastItem);

  return 0;
}

CMsg2::Node *CMsg2::AddItem(unsigned long ulTag)
{
  if(m_nCount>=m_nSize)
  {
    if(m_nGrowBy==0)
      return 0;

    // Grow array
    Node *pNodeArray=new Node[m_nSize+m_nGrowBy];
    memcpy(pNodeArray,m_pNodeArray,sizeof(Node)*m_nSize);
    m_nSize+=m_nGrowBy;
    if(m_pNodeArray!=0)
      delete [] m_pNodeArray;
    m_pNodeArray=pNodeArray;
  }

  m_pNodeArray[m_nCount].m_ulName=ulTag;
  return &m_pNodeArray[m_nCount++];
}

CMsg2::Node::Node()
  :
  m_ulName(0),
  m_eType(CMsg2::TypInvalid),
  m_nArrayCount(0),
  m_Value(0)
{
}

void CMsg2::Node::DeleteValue()
{
  if(m_Value.m_lNumber==0)
    return;

  switch(m_eType)
  {
  case CMsg2::TypString:
    if(m_Value.m_pString!=0)
      delete [] m_Value.m_pString;
    m_Value.m_pString=0;
    break;

  case CMsg2::TypByteArray:
    delete [] m_Value.m_pByteArray;
    m_Value.m_pByteArray=0;
    break;

  case CMsg2::TypWordArray:
    delete [] m_Value.m_pWordArray;
    m_Value.m_pWordArray=0;
    break;

  case CMsg2::TypDwordArray:
    delete [] m_Value.m_pDwordArray;
    m_Value.m_pDwordArray=0;
    break;

  case CMsg2::TypMsg:
    delete m_Value.m_pMsg;
    m_Value.m_pMsg=0;
    break;

  case CMsg2::TypInvalid:
  case CMsg2::TypDword:
  case CMsg2::TypWord:
  case CMsg2::TypByte:
  default:
    // Nothing to delete
    break;
  }

  m_Value.m_lNumber=0;
  m_nArrayCount=0;
  m_eType=CMsg2::TypInvalid;
}

CMsg2::Node::Value::Value(long lValue)
  :
  m_lNumber(lValue)
{
}

// Simple data types
bool CMsg2::SetNumber(const char *pszName,long lNum)
{
  unsigned long ulName=NameToTag(pszName);

  // Find the item that corresponds to the name (if it exists)
  Node *pNode=FindItem(ulName);
  if(pNode==0)
  {
    // Add item
    pNode=AddItem(ulName);
    if(pNode==0)
      return false;
  }
  else
  {
    pNode->DeleteValue();
  }

  pNode->m_eType=TypDword;
  pNode->m_Value.m_lNumber=lNum;
  return true;
}

bool CMsg2::SetShortArray (const char *pszName,unsigned int nItems,short *pData)
{
  return SetWordArray(pszName,nItems,(unsigned short *)pData);
}

bool CMsg2::SetLongArray(const char *pszName,unsigned int nItems,long *pData)
{
  return SetDwordArray(pszName,nItems,(unsigned long *)pData);
}

char *DwordToName(unsigned long ulName,char *pszName2,const char *pszPrefix)
{
  if((ulName&(~0x1FFFFF))!=0)   
    return 0;   // Number is too large to encode

  unsigned char *pszName=(unsigned char *)pszName2;

  for(int i=0; i<4; i++)
  {
    if(pszPrefix!=0 && *pszPrefix!='\0')
    {
      pszName[i]=*pszPrefix;
      pszPrefix++;
    }
    else
    {
      pszName[i]=(unsigned char)((ulName&0x7F) | 0x80);
  ulName>>=7;
    }
  }

  pszName[4]='\0';

  return pszName2;
}

unsigned long NameToTag(const char *pszName2)
{
  unsigned long ulName=0;
  const unsigned char *pszName=(unsigned char *)pszName2;

  if(pszName!=0 && *pszName!='\0')
  {
    ulName=*pszName++;

    ulName<<=8;
    if(*pszName!='\0')
      ulName|=*pszName++;

    ulName<<=8;
    if(*pszName!='\0')
      ulName|=*pszName++;

    ulName<<=8;
    if(*pszName!='\0')
      ulName|=*pszName;
  }

  return ulName;
}
